/* elements.h -- definitions for typesetting elements
 * $Id: elements.h,v 1.7 92/09/18 17:56:25 connolly Exp $
 * Copyright 1992 by Convex Computer Corporation, Richardson, Texas.
 * See the COPYRIGHT file for details.
 */

#ifndef __elements_h
#define __elements_h

#include "space.h"
#include "textflow.h"

class View;

class ParagraphFormat{
public:
  ParagraphFormat();

  enum {max_tabs = 10};
  enum {quad_left, quad_center, quad_right, quad_full} justification;

  Inches left_indent;
  Inches right_indent;
  Inches first_indent;
  Inches space_before;
  Inches space_after;
  Inches space_between;
  char border_top, border_bottom;

  void normal();

  void add_tab_stop(Inches ts)
    { if(tabqty < max_tabs)
	stops[tabqty++] = ts;
    };

  int     tabs() const
    { return tabqty; }

  void    newtabs()
    { tabqty=0; }
  
  static void default_tab_width(Inches w) { default_tab_width_ = w; };
  // This must be set before any calls to tab() are made.

  Inches tab(int n) const
    // tab(3) returns 2" if there are only 3 tab stops
    { return n < tabqty ? stops[n] : (double)(n+1) * default_tab_width_; };

private:
  Inches stops[max_tabs];
  int tabqty;
  static Inches default_tab_width_;
};


class CharacterFormat{
public:
  CharacterFormat()
    { plain(); };
  void plain()
    { bold = italic = underline = font_family = super = sub = 0;
      foreground = 0;
      size = 24;
    };

  char size;

  char bold, italic, underline;
  char font_family, super, sub;
  int foreground;
  unsigned long fg_pixel;
};


/***** formatting elements ******/

class NewLineFlow : public TextFlow{
public:
  NewLineFlow(TextFlow& parent, TextPosition first, TextPosition last)
    : TextFlow(parent, first, last) {};

  Qty character_shape(TextPosition, Qty, Extent, Extent&, int) const;

  virtual Coord font_ascent() const
    { return 1; }; // must be > 0

  virtual Coord space_after() const
    { return 0; };
};

class ParagraphFlow : public TextFlow{
public:
  ParagraphFlow(TextFlow& p, TextPosition f, TextPosition l,
		const ParagraphFormat& fmt, View& v)
    : TextFlow(p, f, l) { view_ = &v; fmt_ = fmt; };

  void line_shape(TextPosition, Extent&) const;

  const View& view() const
    { return *view_; };

protected:
  View* view_;
  ParagraphFormat fmt_;
};


class TabFlow : public TextFlow{
public:
  TabFlow(TextFlow& p, TextPosition f, TextPosition l,
	  ParagraphFormat& pf, View& v)
    : TextFlow(p, f, l) { view_ = &v; fmt_ = pf; };

  Qty character_shape(TextPosition, Qty, Extent, Extent&, int) const;

  Coord next_tab_stop(Coord here) const;

  const View& view() const
    { return *view_; };

protected:
  View* view_;
  ParagraphFormat fmt_;
};

#endif

