/*
 * image.c
 *
 * $Id: image.c,v 1.5 92/08/20 16:47:39 connolly Exp $
 * Copyright 1992 by Convex Computer Corporation, Richardson, Texas.
 * See the COPYRIGHT file for details.
 *
 * convert a hex-encoded sunraster/RLE bitmap string into a Pixmap
 * should handle pseudocolor/staticcolor to 16 bits
 * should handle truecolor/directcolor to 32 bits
 *
 * Wade Smith Mon Jan 27 16:35:39 CST 1992
 *
 */

#ifndef __lint
static char rcsid[] = "$Id: image.c,v 1.5 92/08/20 16:47:39 connolly Exp $";
#endif

/*
 * $Log:	image.c,v $
 * Revision 1.5  92/08/20  16:47:39  connolly
 * 	added RCS Id keywords and copyrights
 * 
 * Revision 1.4  92/08/14  20:11:23  connolly
 * 	RichText.c: fixed a selection core dump bug
 * 	elements.C: underlining and stuff
 * 	image.c: b/w only again
 * 	rtftoken: eof bug fix
 * 	xtextview: image stuff
 * 
 * Revision 1.3  92/02/12  20:19:50  connolly
 * 	added support for 1-bit images on monochrome displays
 * 
 * Revision 1.1  92/01/27  16:40:46  wsmith
 * Initial revision
 * 
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <sys/param.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#include "rasterfile.h"

static char *endian = "\377\0";
#define	IS_BIG_ENDIAN	(*(short*)endian<0)
#define	IS_LITTLE_ENDIAN	(*(short*)endian>0)

#define	MAXCOLORS	2
#define	RCOL		0
#define	GCOL		1
#define	BCOL		2

typedef struct raster {
	int		width, height;
	int		ncolors;
	unsigned char	map[MAXCOLORS][3]; 
	unsigned char	*image;
} Raster;

typedef struct rasterfile	RF;
static GC	gc;

static unsigned char	tohex[256] = {
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,1,2,3,4,5,6,7,8,9,0,0,0,0,0,0,0,10,11,12,13,14,
15,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,10,11,12,13,14,15,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
0,0,0,0,0,0,0,0,0,
};

/* reverse the bytes in a *(int *) */
static void
reverse_int(int *integer)
{
	char	*p = (char *)integer, tmp;
	int	i, j;

	for (i = 0, j = sizeof(int)-1; i < j; i++, j--) {
		tmp = p[i];
		p[i] = p[j];
		p[j] = tmp;
	}
}

/* return the next decoded byte from a hex string */
static int
getbyte(const char* string)
{
	static unsigned char*	ptr;
	int			i, byte;

	if (string) {
		ptr = (unsigned char *)string;
		return EOF;
	}
	
	byte = 0;
	for (i = 0; i < 2; ptr++)
		if (isxdigit(*ptr))
			byte = (byte<<4)|tohex[*ptr], i++;
		else if (*ptr != '\n') 
			return EOF;

	return byte;
}

/* return a block of characters from a hex string */
static int
bread(char* buf, int nbytes)
{
	int	i, byte;

	for (i = 0; i < nbytes; i++) {
		if ((byte = getbyte(NULL)) == EOF)
			return i ? i : EOF;
		*buf++ = byte;
	}
	return i;
}

/* change byte order for little-endian machines */
static void
ReverseRasterfileHeader(RF *rf)
{
	reverse_int( &rf->ras_magic );
	reverse_int( &rf->ras_width );
	reverse_int( &rf->ras_height );
	reverse_int( &rf->ras_depth );
	reverse_int( &rf->ras_length );
	reverse_int( &rf->ras_type );
	reverse_int( &rf->ras_maptype );
	reverse_int( &rf->ras_maplength );
}

/* extract raster info from hex string */
static Raster *
GetRasterFromString(const char* string)
{
	unsigned char	*src, *dst, *end, *data;
	int	ch, size, bytes_per_line, depth;
	int	x, y, bit;
	RF	rf;
	Raster	raster, *new;

	(void) getbyte(string);

	if (bread((char *)&rf, sizeof(rf)) != sizeof(rf)) return NULL;
	if (IS_LITTLE_ENDIAN) ReverseRasterfileHeader(&rf);
	if (rf.ras_magic != RAS_MAGIC) return NULL;
	if (rf.ras_depth != 1) return NULL;
	if (rf.ras_type != RT_BYTE_ENCODED) return NULL;

	raster.width = rf.ras_width;
	raster.height = rf.ras_height;
	if (raster.width <= 0 || raster.height <= 0) return NULL;
	depth = rf.ras_depth;

	if (depth == 1 && rf.ras_maptype == RMT_NONE) {
		raster.ncolors = 2;
		raster.map[0][RCOL] = 255;	/* slot 0 - white */
		raster.map[0][GCOL] = 255;
		raster.map[0][BCOL] = 255;
		raster.map[1][RCOL] = 0;	/* slot 1 - black */
		raster.map[1][GCOL] = 0;
		raster.map[1][BCOL] = 0;
	} else
		return NULL;

	bytes_per_line = ((raster.width*depth + 15)/16)*2;
	size = bytes_per_line * raster.height;
	raster.image = (unsigned char *)malloc( size );
	if (raster.image == NULL) return NULL;

	dst = raster.image;
	end = raster.image+size;
	while (dst < end) {
		ch = getbyte(NULL);
		if (ch == EOF) { free(raster.image); return NULL; }
		if (ch == 0x80) {
			int n = getbyte(NULL);
			if (n == EOF) { free(raster.image); return NULL; }
			if (n > 0) {
				ch = getbyte(NULL);
				if (ch == EOF) 
					{ free(raster.image); return NULL; }
				while (dst < end && n-- > -1) 
					*dst++ = ch;
			} else
				*dst++ = ch;
		} else
			*dst++ = ch;
	}

	data = (unsigned char *)malloc( raster.width*raster.height );
	if (data == NULL) { free(raster.image); return NULL; }
	/* assume big-endian bit order */
	for (y = 0; y < raster.height; y++) {
	    src = raster.image+y*bytes_per_line;
	    dst = data+y*raster.width;
	    bit = 7;
	    for (x = 0; x < raster.width; x++) {
		*dst++ = (*src>>bit)&1;
		if (--bit < 0) { src++; bit = 7; }
	    }
	}
	free( raster.image );
	raster.image = data;

	new = (Raster *)malloc( sizeof(raster) );
	if (new == NULL) return NULL;

	(void) memcpy( (void *)new, (void *)&raster, sizeof(raster) );
	return new;
}

/* convert bitmap image to XImage format */
static XImage *
create_image(Display* display, Visual* visual, int depth, Raster* raster)
{
	char	*data;
	int	i, size, bits_per_pixel, width;
	unsigned long	map[2];
	XImage	*image;

	map[0] = WhitePixel(display, DefaultScreen(display));
	map[1] = BlackPixel(display, DefaultScreen(display));

	width = raster->width;
	size = width * raster->height;

	if (depth > 16) {
		unsigned long	*ptr;
		ptr = (unsigned long *)malloc( size * sizeof(long) );
		if (ptr == NULL) return NULL;
		for (i = 0; i < size; i++)
			ptr[i] = (unsigned long)map[ raster->image[i] ];
		data = (char *)ptr;
		bits_per_pixel = 32;
	} else if (depth > 8) {
		unsigned short	*ptr;
		ptr = (unsigned short *)malloc( size * sizeof(short) );
		if (ptr == NULL) return NULL;
		for (i = 0; i < size; i++)
			ptr[i] = (unsigned short)map[ raster->image[i] ];
		data = (char *)ptr;
		bits_per_pixel = 16;
	} else if (depth > 1) {
		unsigned char	*ptr;
		ptr = (unsigned char *)malloc( size * sizeof(char) );
		if (ptr == NULL) return NULL;
		for (i = 0; i < size; i++)
			ptr[i] = (unsigned char)map[ raster->image[i] ];
		data = (char *)ptr;
		bits_per_pixel = 8;
	} else {
		unsigned char	*ptr, *img;
		int		x,y;
		width = (width+7)&~7;
		size = width * raster->height;
		ptr = (unsigned char *)(data = calloc( size/8, 1 ));
		if (ptr == NULL) return NULL;
		img = raster->image;
		for (y = 0; y < raster->height; y++, ptr+=width/8)
			for (x = 0; x < raster->width; x++)
				ptr[x/8] |= (unsigned char)(map[*img++]<<(x&7));
		bits_per_pixel = 1;
	}

	image = XCreateImage(display, visual, depth, ZPixmap, 0, (char *)data, 
			     raster->width, raster->height, bits_per_pixel,
			     (width*bits_per_pixel)/8);
	if (image == NULL) {
		free(data);
		return NULL;
	}
	
	image->byte_order = (IS_BIG_ENDIAN ? MSBFirst : LSBFirst);
	image->bitmap_unit = bits_per_pixel;
	image->bitmap_bit_order = LSBFirst;

	image->bits_per_pixel = bits_per_pixel;
	return image;
}

/* free Raster* struct */
static void
FreeRaster(Raster* raster)
{
	free( (char *)raster->image );
	free( (char *)raster );
}

static void
initializeImageBuilder(Display* display)
{
	Window root = DefaultRootWindow(display);
	gc = XCreateGC(display, root, 0L, (XGCValues *)0);
}

Pixmap
createPixmapFromHexData(Display* display, Drawable drawable, const char* data)
/* USE:
 *	data = "0123456789ABCDEF\nFF00FF00FF00FF00\n...";
 *	pixmap = createPixmapFromHexData(disp, drawable, data);
 */
{
	int screen = DefaultScreen(display);
	Visual* visual = DefaultVisual(display,screen);
	int depth  = DefaultDepth(display,screen);
	Raster	*raster;
	Pixmap	pixmap;
	XImage	*image;
	static int inited=0;

	if(!inited){
	  initializeImageBuilder(display);
	  inited = 1;
	}

	if (!(visual->class == PseudoColor || visual->class == StaticColor ||
	    visual->class == DirectColor || visual->class == TrueColor ||
	    visual->class == StaticGray))
	  return 0;

	raster = GetRasterFromString(data);
	if (raster == NULL) return 0;

	image = create_image(display, visual, depth, raster);

	if (image == NULL) {
		FreeRaster(raster);
		return 0;
	}

	pixmap = XCreatePixmap(display, drawable, 
			       raster->width, raster->height, depth);
	if (!pixmap) {
		FreeRaster(raster);
		XDestroyImage(image);
		return 0;
	}

	XPutImage(display, pixmap, gc, image, 
		  0, 0, 0, 0, raster->width, raster->height);

	FreeRaster(raster);
	XDestroyImage(image);
	return pixmap;
}
