/* rt_test.c -- test driver for RichText widget.
 * $Id: rt_test.c,v 1.6 92/09/22 18:14:01 connolly Exp $
 * Copyright 1992 by Convex Computer Corporation, Richardson, Texas.
 * See the COPYRIGHT file for details.
 */


#include <Xm/Xm.h>
#include <Xm/ScrolledW.h>
#include <Xm/MainW.h>
#include <Xm/FileSB.h>
#include <Xm/PushB.h>

#include "RichText.h"
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>

static void
clickCB(Widget caller, XtPointer dummy, XtPointer cbd)
{
  XcRichTextCallbackStruct* rt = (XcRichTextCallbackStruct*)cbd;

  printf("selection: %.*s\n", rt->point - rt->mark, rt->value + rt->mark);
  printf("next 10 chars: %.10s\n", rt->value + rt->point);

  if(rt->point == rt->mark){
	Boolean rtf;
	XtVaGetValues(caller, XcNrtf, &rtf, NULL);
	printf("RichText interpretation: %s\n", rtf ? "off" : "on");
	XtVaSetValues(caller, XcNrtf, !rtf, NULL);
  }
}


static void resizeHandler(Widget w, XtPointer pt, XtPointer motifdata)
{
  Widget rt = (Widget)pt;
  Dimension width;

  XtVaGetValues(w,
		XmNwidth, &width,
		NULL);

  /* ignore sizes that are way too small */
  if(width > 100)
    XtVaSetValues(rt,
		  XmNwidth, width - 15,
		  NULL);
}

static void showfile(Widget w, XtPointer pt, XtPointer md)
{
  Widget rt = (Widget)pt;
  XmFileSelectionBoxCallbackStruct* fs = (XmFileSelectionBoxCallbackStruct*)md;
  FILE* f;
  struct stat s;
  char* text;
  char* fname;
  int bottom;
  char* page_end;

  if(!XmStringGetLtoR(fs->value, XmSTRING_DEFAULT_CHARSET, &fname)){
    XtWarning("can't XmStringGetLtoR");
    return;
  }

  if(stat(fname, &s)){
    XtWarning("can't stat file");
    XtFree(fname);
    return;
  }

  f = fopen(fname, "r");
  XtFree(fname);

  if(!f){
    XtWarning("%s: can't open %s\n");
    return;
  }

  if(!(text = (char*)XtMalloc(s.st_size + 1))){
    XtWarning("%s: can't malloc enough memory for: %s\n");
    return;
  }
  
  if(fread(text, sizeof(char), s.st_size, f) != s.st_size){
    XtWarning("%s: error reading for: %s\n");
    XtFree(text);
    return;
  }

  fclose(f);
  text[s.st_size] = 0;

  for(page_end = strstr(text, "\\page");
      page_end && isalpha(*(page_end + 5));
      page_end = strstr(page_end+1, "\\page"))
    {}

  bottom = page_end ? (page_end - text) : 0;

  XtVaSetValues(rt,
		XcNrtf, text[0] == '{', /* cheap hack test */
		XmNvalue, text,
		XmNbottomPosition, bottom,
		NULL);
}


#ifdef SEL
static void rangeCB(Widget w, XtPointer pt, XtPointer md)
{
  Widget rt = (Widget)pt;
  XmSelectionBoxCallbackStruct* fs = (XmSelectionBoxCallbackStruct*)md;
  char* text;
  XmTextPosition start, end;

  if(!XmStringGetLtoR(fs->value, XmSTRING_DEFAULT_CHARSET, &text)){
    XtWarning("can't XmStringGetLtoR");
    return;
  }

  if(sscanf(text, "%d %d", &start, &end) != 2){
    XtWarning("type <start> <end> into the selection area");
    XtFree(text);
    return;
  }

  XtFree(text);

  XtVaSetValues(rt,
		XmNtopPosition, start,
		XmNbottomPosition, end,
		NULL);
}
#endif

static void pageCB(Widget w, XtPointer pt, XtPointer md)
{
  Widget rt = (Widget)pt;
  char* text;
  int start, end;
  char* page_end;

  XtVaGetValues(rt,
		XmNvalue, &text,
		XmNtopPosition, &start,
		XmNbottomPosition, &end,
		NULL);

  for(page_end = strstr(text+end+1, "\\page");
      page_end && isalpha(*(page_end + 5));
      page_end = strstr(page_end+1, "\\page"))
    {}

  if(page_end){
    start =  end;
    end = page_end - text;

    XtVaSetValues(rt,
		  XmNtopPosition, start,
		  XmNbottomPosition, end,
		  NULL);
  }
}

int
main(argc, argv)
Cardinal argc; char *argv[];
{
  XtAppContext app;
  Widget appShell;
  Widget rt, sc, clip, fsb, pager;
#ifdef SEL
  Widget sel;
#endif

  appShell = XtAppInitialize(&app, "RichTextTest",
			     NULL, 0,
			     &argc, argv, NULL,
			     (ArgList)NULL, 0);

#ifdef SEL
  sel = XmCreateSelectionDialog(appShell, "range selection",
				0, 0);
  XtManageChild(sel);
#endif

  sc = XtVaCreateManagedWidget("scroller", xmMainWindowWidgetClass,
			       appShell,
			       XmNscrollBarDisplayPolicy, XmAS_NEEDED,
			       XmNscrollingPolicy, XmAUTOMATIC,
			       NULL);

  fsb = XmCreateFileSelectionBox(sc, "filechooser",
				    0, 0);
  XtManageChild(fsb);

  rt = XtVaCreateManagedWidget("richtext", xcRichTextWidgetClass,
			       sc,
			       NULL);

  pager = XmCreatePushButton(sc, "pager", NULL, 0);

  XtManageChild(pager);

  XmMainWindowSetAreas(sc, pager, fsb, NULL, NULL, rt);

  /* In order to allow the user to change the width
     of the RichText widget, I trap resize events
     on the clip window. */
  XtVaGetValues(sc,
		XmNclipWindow, &clip,
		NULL);
  XtAddCallback(clip, XmNresizeCallback, resizeHandler, (XtPointer)rt);


  /* show files on request */
  XtAddCallback(fsb, XmNokCallback, showfile, rt);

  /* provide some feedback to clicks in the text area */
  XtAddCallback(rt, XmNmotionVerifyCallback, clickCB, 0);

#ifdef SEL
  /* change range of richtext widget */
  XtAddCallback(sel, XmNapplyCallback, rangeCB, rt);
#endif

  /* next \page */
  XtAddCallback(pager, XmNactivateCallback, pageCB, rt);

  XtRealizeWidget ( appShell );
  XtAppMainLoop ( app );
  
  return 0;
}
